# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2007 Async Open Source
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public License
## as published by the Free Software Foundation; either version 2
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

import datetime
from decimal import Decimal

from . import latscii
latscii.register_codec()

number = (int, Decimal)


def argtype_name(argtype):
    if argtype == number:
        return 'number'
    else:
        return argtype.__name__


class SintegraError(Exception):
    pass


class SintegraFile(object):
    def __init__(self):
        self._registers = []

    def add(self, register):
        """Adds a register to the file
        :param register: a register
        :type register: :class:`SintegraRegister`
        """
        if not isinstance(register, SintegraRegister):
            raise TypeError("register must be a SintegraRegister instance")

        numbers = [f.sintegra_number for f in self._registers]
        if register.sintegra_unique:
            if register.sintegra_number in numbers:
                raise SintegraError("%s can only be added once" % (register.sintegra_number, ))
        if register.sintegra_requires:
            for number in register.sintegra_requires:
                if not number in numbers:
                    raise SintegraError("%s must be added at this point" % (number, ))

        self._registers.append(register)

    def add_header(self, cgc, estadual, company, city, state, fax, start, end):
        """Receive values to generate Sintegra Register type 10.

        :param cgc: the branch CNPJ number.
        :param estadual: the branch 'Inscrição estadual' number or ISENTO.
        :param company: the company fancy name.
        :param city: the branch city.
        :param state: the branch city state.
        :param fax: the branch fax number.
        :param start: start's date period, generally, the 1th month day.
        :type start: datetime.date
        :param end: end's date periodo, generally, the last month day.
        :type end: datetime.date
        """
        self.add(SintegraRegister10(
            cgc, estadual, company,
            city, state, fax,
            int(start.strftime('%Y%m%d')),
            int(end.strftime('%Y%m%d')),
            '331'))

    def add_complement_header(self, address, number, complement, district,
                              postal, name, phone):
        """Receive values to generate Sintegra Register type 11.

        :param address: the branch address.
        :param number: the number of the branch address.
        :param complement: the complement of the branch address.
        :param district: district of the branch address.
        :param postal: postal code number of the branch address.
        :param name: the branch manager name.
        :param phone: the branch phone number.
        """
        self.add(SintegraRegister11(
            address, number, complement, district,
            postal, name, phone))

    def add_fiscal_coupon(self, date, printerserial, printerid,
                          coupon_start, coupon_end, crz, cro, period_total,
                          total):
        """Receive values for generate 60M Sintegra Register.

        :param date: emission date of the fiscal coupon.
        :type date: datetime.date
        :param printerserial: serial number of the fiscal printer.
        :param printerid: the refered number (id) for the fiscal printer
          in a branch.
        :param coupon_start: the number in which the coupon fiscal starts.
        :param coupon_end: the number in which the fiscal coupon ends.
        :param crz: counter the number of 'Zs reduction' made by fiscal printer.
        :param cro: counter how many times the fiscal printer was restarted
          their operations.
        :param period_total: value total in a fiscal day.
        :param total: total acumulated in fiscal printer.
        """
        if not isinstance(date, datetime.date):
            raise TypeError
        idate = int(date.strftime('%Y%m%d'))
        self.add(SintegraRegister60M('M', idate, printerserial, printerid, '2D',
                                     coupon_start, coupon_end, crz, cro,
                                     period_total,
                                     total))

    def add_fiscal_tax(self, date, printerserial, code, value):
        """
        Receive values for generate 60A Sintegra Register.
        :param date: emission date of the fiscal coupon.
        :type date: datetime.date
        :param printerserial: serial number of the fiscal printer.
        :param code: the tax code.
        :param value: the tax value.
        """
        self.add(SintegraRegister60A(
            'A', int(date.strftime('%Y%m%d')), printerserial,
            code,
            int(value * 100)))

    def add_products_summarized(self, date, product_code,
                                product_quantity,
                                total_liquido_produto,
                                total_icms_base,
                                icms_aliquota):
        code = "%014s" % product_code
        product_code = code.replace(' ', '0')
        icms_aliquota = '%04d' % (icms_aliquota)
        if icms_aliquota == '0000':
            icms_aliquota = "I"
        self.add(SintegraRegister60R('R', date, product_code,
                                     product_quantity,
                                     total_liquido_produto,
                                     total_icms_base,
                                     icms_aliquota, ""))

    def add_receiving_order(self, cnpj, state_registry, receival_date,
                            state, modelo, serial, numero, cfop, emitente,
                            total, icms_base, icms_total, isenta, outras,
                            aliquota_icms, situacao):
        cfop_code = cfop.replace(".", "")
        date = int(receival_date.strftime("%Y%m%d"))
        self.add(SintegraRegister50(cnpj, str(state_registry),
                                    date,
                                    str(state), modelo, serial, numero,
                                    int(cfop_code),
                                    emitente,
                                    total * 100,
                                    icms_base * 100,
                                    icms_total * 100,
                                    isenta * 100,
                                    outras * 100,
                                    aliquota_icms * 100,
                                    situacao))

    def add_receiving_order_item(self, cnpj, modelo, serial, numero, cfop, cst,
                                 numero_item, product_code, product_quantity,
                                 valor_bruto_produto, desconto, icms_base,
                                 icms_subst_trib, ipi, icms_aliquota):
        cfop_code = cfop.replace(".", "")
        cfop_int = int(cfop_code)
        if product_code is None:
            product_code = " " * 14
        else:
            code = '%014s' % (product_code, )
            product_code = code.replace(' ', '0')
        if cst is None:
            cst = ' ' * 3
        self.add(SintegraRegister54(cnpj, modelo, serial, numero, cfop_int,
                                    cst, numero_item, product_code,
                                    product_quantity * 1000,
                                    valor_bruto_produto * 100,
                                    desconto * 100,
                                    icms_base * 100,
                                    icms_subst_trib * 100,
                                    ipi * 100,
                                    icms_aliquota * 100))

    def add_inventory_item(self, start, product_code, product_quantity,
                           total_product_value, owner_code, owner_cnpj,
                           owner_state_registry, state):
        inventory_date = int(start.strftime("%Y%m%d"))
        code = '%014s' % product_code
        product_code = code.replace(' ', '0')
        blank = ' ' * 45
        # When the actual company owns the products, the owner_cnpj and
        # owner_state_registry fields must be filled with pre-defined values.
        # See the link in #3708 for details.
        if owner_code == 1:
            owner_cnpj = 0
            owner_state_registry = ' ' * 14

        self.add(SintegraRegister74(inventory_date,
                                    product_code,
                                    product_quantity,
                                    total_product_value,
                                    owner_code,
                                    owner_cnpj,
                                    owner_state_registry,
                                    state,
                                    blank))

    def add_product(self, start, end, product_code, ncm, desc,
                    unit, aliquota_ipi, aliquota_icms,
                    reducao_icms, base_icms):
        start = int(start.strftime("%Y%m%d"))
        end = int(end.strftime("%Y%m%d"))
        product_code = '%-14s' % product_code
        ncm = '%08d' % (int(ncm))

        self.add(SintegraRegister75(start, end, product_code, ncm, desc,
                                    unit, aliquota_ipi, aliquota_icms,
                                    reducao_icms, base_icms))

    def close(self):
        """Closes the file.
        This will add a couple of registers of type 90.
        """
        sums = {}
        for register in self._registers[2:]:
            number = register.sintegra_number
            if not number in sums:
                sums[number] = 1
            else:
                sums[number] += 1
        cgc = self._registers[0].cgc
        estadual = self._registers[0].estadual
        totalizers = len(sums) + 1
        for number, fsum in sorted(sums.items()):
            self.add(SintegraRegister90(cgc, estadual, number, fsum, '',
                                        totalizers))
        self.add(SintegraRegister90(cgc, estadual, 99,
                                 len(self._registers) + 1, '', totalizers))

    def write(self, filename=None, fp=None):
        """Writes out of the content of the file to a filename or fp

        :param filename: filename
        :param fp: file object, anything implementing write(data)
        """
        if filename is None and fp is None:
            raise TypeError
        if filename is not None and fp is not None:
            raise TypeError
        if fp is None:
            fp = open(filename, 'wb')

        for register in self.get_registers():
            fp.write(register.get_string())

    def get_registers(self):
        last_register = self._registers[-1]
        if (last_register.sintegra_number != 90 or
            last_register.type != 99):
            raise TypeError("You need to close the document before calling write()")
        return self._registers


class SintegraRegister(object):
    """ This is an abstract class
    The arguments depends on what is defined in the class variable
    sintegra_fields

    :cvar sintegra_number:
    :cvar sintegra_fields:
    :cvar sintegra_unique:
    :cvar sintegra_requires:
    """

    sintegra_number = 0
    sintegra_fields = None
    sintegra_unique = False
    sintegra_requires = None

    def __init__(self, *args, **kwargs):
        """ Creates a new SintegraRegister"""
        if not self.sintegra_fields:
            raise TypeError
        if not self.sintegra_number:
            raise TypeError
        if len(args) != len(self.sintegra_fields):
            raise TypeError('%s expected %d parameters but got %d' % (
                self.__class__.__name__, len(self.sintegra_fields), len(args)))

        sent_args = dict([(field[0], field[1:2])
                          for field in zip(self.sintegra_fields, args)])
        for key in kwargs:
            if key in sent_args:
                raise SintegraRegister("%s specified two times" % (key, ))

        total = 0
        self._values = {}
        for (name, length, argtype), arg in zip(self.sintegra_fields, args):
            if arg is None:
                pass
            elif not isinstance(arg, argtype):
                raise TypeError(
                    "argument %s should be of type %s but got %s" % (
                    name, argtype_name(argtype), type(arg).__name__))
            self._values[name] = self._arg_to_string(arg, length, argtype)
            setattr(self, name, arg)
            total += length

        if total > 124:
            raise TypeError(
                "There are items with a total length of %d in %s, "
                "but only 124 is allowed" % (total, self.__class__.__name__))
        elif total < 124:
            self.padding = 124 - total
        else:
            self.padding = 0

    #
    # Public API
    #

    def get_string(self):
        """
        Gets a string for all sintegra fields.
        :returns: sintegra fields as string.
        """
        values = []
        for (name, _, _) in self.sintegra_fields:
            values.append(self._values[name])
        if self.padding:
            values.append(' ' * self.padding)
        return '%02d%s\r\n' % (self.sintegra_number,
                              ''.join(values))

    # Private

    def _arg_to_string(self, value, length, argtype):
        if value is None:
            return ' ' * length

        if argtype == number:
            # If a value is higher the the maximum allowed,
            # set it to the maximum allowed value instead.
            max_value = (10 ** length) - 1
            if value > max_value:
                value = max_value
            arg = '%0*d' % (length, value)
        elif argtype == str:
            # Accept normal strings, which are assumed to be UTF-8
            if type(value) == str:
                value = str(value.encode(), 'utf-8')

            # Convert to latscii
            value = value.encode('ascii', 'replacelatscii').decode()

            arg = '%-*s' % (length, value)
            # Chop strings which are too long
            if len(arg) > length:
                arg = arg[:length]
        else:
            raise AssertionError
        assert len(arg) <= length
        return arg


class SintegraRegister10(SintegraRegister):
    sintegra_number = 10
    sintegra_fields = [
        ('cgc', 14, number),
        ('estadual', 14, str),
        ('company', 35, str),
        ('city', 30, str),
        ('state', 2, str),
        ('fax', 10, number),
        ('start_date', 8, number),
        ('end_date', 8, number),
        ('codes', 3, str),
        # 1: 1..3
        # 2: 1..3
        # 3: 1..3,5
        ]
    sintegra_unique = True


class SintegraRegister11(SintegraRegister):
    sintegra_number = 11
    sintegra_fields = [
        ('place', 34, str),
        ('number', 5, number),
        ('complement', 22, str),
        ('dibasestringict', 15, str),
        ('postal', 8, number),
        ('name', 28, str),
        ('phone', 12, number),
        ]

    sintegra_requires = 10,
    sintegra_unique = True


class SintegraRegister60M(SintegraRegister):
    sintegra_number = 60
    sintegra_fields = [
        ('type', 1, str),
        ('date', 8, number),
        ('printerserial', 20, str),
        ('printerid', 3, number),
        ('model', 2, str),
        ('start_coo', 6, number),
        ('end_coo', 6, number),
        ('crz', 6, number),
        ('cro', 3, number),
        ('period_total', 16, number),
        ('total', 16, number),
        ]
    sintegra_requires = 10, 11


class SintegraRegister60A(SintegraRegister):
    sintegra_number = 60
    sintegra_fields = [
        ('type', 1, str),
        ('date', 8, number),
        ('printerserial', 20, str),
        ('tax', 4, str),
        ('total', 12, number),
        ]
    sintegra_requires = 10, 11


class SintegraRegister60D(SintegraRegister):
    sintegra_number = 60
    sintegra_fields = [
        ('type', 1, str),
        ('date', 8, number),
        ('serial_number', 20, str),
        ('product_code', 14, str),
        ('quantity', 13, number),
        ('product_value', 16, number),
        ('total_icms_base', 16, number),
        ('id_situacao_trib', 4, str),
        ('valor_icms', 13, number),
        ('blank', 13, str)
    ]


class SintegraRegister60I(SintegraRegister):
    sintegra_number = 60                       #  1) Tipo 60
    sintegra_fields = [
        ('type', 1, str),               #  2) Subtipo 'I'
        ('date', 8, number),                   #  3) Data de emissão do documento fiscal
        ('printerserial', 20, str),     #  4) Número de série de fabricação do equipamento
        ('model', 2, str),              #  5) Código do modelo do documento fiscal
        ('coo', 6, number),                    #  6) Número do Contador de Ordem de Operação (COO)
        ('product_number', 3, number),         #  7) Número de Ordem do item no Documento Fiscal
        ('product_code', 14, str),      #  8) Código da mercadoria/produto ou serviço do informante
        ('produto_quantity', 13, number),      #  9) Quantidade da mercadoria/produto (com 3 decimais)
        ('product_value', 13, number),         # 10) Vr líq. (vr bruto diminuído do desc.) da merc./prod. (2 decimais)
        ('total_icms_base', 12, number),       # 11) Base de Cálculo do ICMS do Item (com 2 decimais)
        ('id_situacao_trib', 4, str),   # 12) Identificador da Sit. Tributária / Alíquota do ICMS (2 decimais)
        ('icms_aliquota', 12, number),         # 13) Montante do Imposto (2 decimais)
        ('blank', 16, str)
    ]


class SintegraRegister60R(SintegraRegister):
    sintegra_number = 60
    sintegra_fields = [
        ('type', 1, str),
        ('date', 6, number),
        ('product_code', 14, str),
        ('product_quantity', 13, number),
        ('total_liquido_produto', 16, number),
        ('total_icms_base', 16, number),
        ('icms_aliquota', 4, str),
        ('blank', 54, str),
    ]


class SintegraRegister61(SintegraRegister):
    sintegra_number = 61
    sintegra_fields = [
        # ('type', 1, str),
        # 02 ) espaço em branco
        ('branco', 28, str),

        # 04) Data de emissão do(s)  documento(s) fiscal(is)
        ('data_emissao', 8, number),

        # 05) Modelo do(s) documento(s)  fiscal(is)
        ('modelo', 2, number),

        # 06) Série do(s) documento(s)  fiscal(is)
        ('serie', 3, str),

        #07) Subsérie do(s) documento(s) fiscal(is)
        ('subserie', 2, str),

        # 08)Número do primeiro         documento fiscal emitido no         dia do mesmo modelo, série         e subsérie
        ('inicial_ordem', 6, number),

        # 09) Número do último documento  fiscal emitido no dia do mesmo modelo, série e subsérie
        ('final_ordem', 6, number),

        #10)Valor total do(s) documento(s) fiscal(is)/Movimento diário   (com 2 decimais)
        ('valor_total', 13, number),

        # 11) Base de cálculo do ICMS (com duas decimais)
        ('icms_base', 13, number),
        # 12) Valor do Montante do  Imposto/Total diário (com 2  decimais)
        ('icms_valor', 12, number),

        # 13) Valor amparado por isenção ou não incidência (com duas decimais)
        ('isento', 13, number),

        # 14) Valor que não confira débito  ou crédito de ICMS/Total    diário (com 2 decimais)
        ('outras', 13, number),

        #15) Alíquota do ICMS  (com 2 decimais)
        ('aliquota_icms', 4, number),
        #16) em branco
        ('blank', 1, str),

        ]


class SintegraRegister61R(SintegraRegister):
    sintegra_number = 61
    sintegra_fields = [
        ('subtipo', 1, str),
        ('mes_ano_emissao', 6, number),
        ('codigo_produto', 14, str),
        ('quantidade', 13, number),
        ('valor_bruto', 16, number),
        ('base_calculo_icms', 16, number),
        ('aliquota_icms', 4, number),
        ('branco', 54, str)
    ]


class SintegraRegister50(SintegraRegister):
    sintegra_number = 50
    sintegra_fields = [
        ('cnpj', 14, number),
        ('estadual', 14, str),
        ('date', 8, number),
        ('state', 2, str),
        ('modelo', 2, number),
        ('serial', 3, str),
        ('numero', 6, number),
        ('cfop', 4, number),
        ('emitente', 1, str),
        ('total', 13, number),
        ('icms_base', 13, number),
        ('icms_total', 13, number),
        ('isenta', 13, number),
        ('outras', 13, number),
        ('aliquota_icms', 4, number),
        ('situacao', 1, str),
    ]
##
# Registro Tipo 51 - TOTAL DE NOTA FISCAL QUANTO AO IPI
#
class SintegraRegister51(SintegraRegister):
    sintegra_number = 51
    sintegra_fields = [
        ('cnpj', 14, number),           #  2) CNPJ do remetente nas entradas e do destinatário nas saídas
        ('estadual', 14, str),   #  3) Inscrição Estadual do remetente nas entradas e do destinatário nas saídas
        ('date', 8, number),            #  4) Data de emissão na saída ou recebimento na entrada
        ('state', 2, str),       #  5) Sigla da unid. da Fed. do remetente nas entradas e do dest. nas saídas
        ('serial', 3, str),      #  6) Série da nota fiscal
        ('numero', 6, number),          #  7) Número da nota fiscal
        ('cfop', 4, number),            #  8) Código Fiscal de Operação e Prestação
        ('total', 13, number),          #  9) Valor total da nota fiscal (com 2 decimais)
        ('ipi_total', 13, number),      # 10) Montante do IPI (com 2 decimais)
        ('isenta', 13, number),         # 11) Valor amparado por isenção ou não incidência do IPI (com 2 decimais)
        ('outras', 13, number),         # 12) Valor que não confira débito ou crédito do IPI (com 2 decimais)
        ('blank', 20, str),      # 13) Brancos
        ('situacao', 1, str),    # 14) Situação da Nota Fiscal
    ]
    # 11.1.14 - CAMPO 14 - Preencher o campo de acordo com a tabela abaixo:
    # =====================================================================
    # [N] Documento Fiscal Normal
    # [S] Documento Fiscal Cancelado
    # [E] Lançamento Extemporâneo de Documento Fiscal Normal
    # [X] Lançamento Extemporâneo de Documento Fiscal Cancelado
    # [2] Documento com USO DENEGADO - exclusivamente para uso dos emitentes de Nota Fiscal Eletrônica - Modelo 55 e
    #     Conhecimento de Transporte Eletrônico, Modelo 57
    # [4] Documento com USO inutilizado - exclusivamente para uso dos emitentes de Nota Fiscal Eletrônica - Modelo 55 e
    #     Conhecimento de Transporte Eletrônico, Modelo 57

##
# Registro Tipo 53 - SUBSTITUIÇÃO TRIBUTÁRIA
#
class SintegraRegister53(SintegraRegister):
    sintegra_number = 53
    sintegra_fields = [
        ('cnpj', 14, number),           #  2) CNPJ do remetente nas entradas e do destinatário nas saídas
        ('estadual', 14, str),   #  3) Inscrição Estadual do remetente nas entradas e do destinatário nas saídas
        ('date', 8, number),            #  4) Data de emissão na saída ou recebimento na entrada
        ('state', 2, str),       #  5) Sigla da unid. da Fed. do remetente nas entradas e do dest. nas saídas
        ('modelo', 2, number),          #  6) Código do modelo da nota fiscal
        ('serial', 3, str),      #  7) Série da nota fiscal
        ('numero', 6, number),          #  8) Número da nota fiscal
        ('cfop', 4, number),            #  9) Código Fiscal de Operação e Prestação
        ('emitente', 1, str),    # 10) Emitente da Nota Fiscal (P-próprio/T-terceiros)
        ('icms_base', 13, number),      # 11) Base de cálculo de retenção do ICMS (com 2 decimais)
        ('icms_retido', 13, number),    # 12) ICMS retido pelo substituto (com 2 decimais)
        ('despesas', 13, number),       # 13) Soma das despesas acessórias (frete, seguro e outras - com 2 decimais)
        ('situacao', 1, str),    # 14) Situação da Nota Fiscal
        ('cod_antecipacao', 1, str), # 15) Código que identifica o tipo da Antecipação Tributária
        ('blank', 29, str),      # 15) Brancos
    ]
    # 11.1.14 - CAMPO 17 - Preencher o campo de acordo com a tabela abaixo:
    # =====================================================================
    # [N] Documento Fiscal Normal
    # [S] Documento Fiscal Cancelado
    # [E] Lançamento Extemporâneo de Documento Fiscal Normal
    # [X] Lançamento Extemporâneo de Documento Fiscal Cancelado
    # [2] Documento com USO DENEGADO - exclusivamente para uso dos emitentes de Nota Fiscal Eletrônica - Modelo 55 e
    #     Conhecimento de Transporte Eletrônico, Modelo 57
    # [4] Documento com USO inutilizado - exclusivamente para uso dos emitentes de Nota Fiscal Eletrônica - Modelo 55 e
    #     Conhecimento de Transporte Eletrônico, Modelo 57
    #
    # 13.1.9 - CAMPO 15 - Preencher o campo de acordo com a tabela abaixo:
    # ====================================================================
    # [1] Pagamento de substituição efetuada pelo destinatário, quando não efetuada ou efetuada a menor pelo substituto
    # [2] Antecipação tributária efetuada pelo destinatário apenas com complementação do diferencial de aliquot
    # [3] Antecipação tributária com MVA (Margem de Valor Agregado), efetuada pelo destinatário sem encerrar a fase de
    #     tributação
    # [4] Antecipação tributária com MVA (Margem de Valor Agregado), efetuada pelo destinatário encerrando a fase de
    #     tributação
    # [5] Substituição tributária interna motivada por regime especial de tributação
    # [6] ICMS pago na importação
    # [ ] Substituição Tributária informada pelo substituto ou pelo substituído que não incorra em nenhuma das situações
    #     anteriores
class SintegraRegister54(SintegraRegister):
    sintegra_number = 54
    sintegra_fields = [
        ('cnpj', 14, number),
        ('modelo', 2, number),
        ('serial', 3, str),
        ('numero', 6, number),
        ('cfop', 4, number),
        ('cst', 3, str),
        ('numero_item', 3, number),
        ('product_code', 14, str),
        ('product_quantity', 11, number),
        ('valor_bruto_produto', 12, number),
        ('desconto', 12, number),
        ('icms_base', 12, number),
        ('icms_subst_trib', 12, number),
        ('ipi', 12, number),
        ('icms_aliquota', 4, number),
        ]

##
# Registro Tipo 55 - GUIA NACIONAL DE RECOLHIMENTO DE TRIBUTOS ESTADUAIS
#
class SintegraRegister55(SintegraRegister):
    sintegra_number = 55
    sintegra_fields = [
        ('cnpj', 14, number),
        # 2) CNPJ do remetente nas entradas e do destinatário nas saídas
        ('estadual', 14, str),
        # 3) Inscrição Estadual do remetente nas entradas e do destinatário nas saídas
        ('date', 8, number),
        # 4) Data do pagamento do documento de Arrecadação
        ('state', 2, str),
        # 5) Sigla da unidade da Federação do contribuinte substituto tributário
        ('state_dest', 2, str),
        # 6) Sigla da unidade da Federação de destino (favorecida)
        ('banco', 3, number),
        # 7) Código do Banco onde foi Efetuado o recolhimento
        ('agencia', 4, number),
        # 8) Agência onde foi efetuado o recolhimento
        ('numero', 20, str),
        # 9) Número de autenticação Bancária do documento de Arrecadação
        ('valor', 13, number),
        # 10) Valor recolhido (com 2 decimais)
        ('vencimento', 8, number),
        # 11) Data do vencimento do ICMS substituído
        ('mes_ano', 6, number),
        # 12) Mês e ano referente à ocorrência do fato gerador, formato MMAAAA
        ('convenio', 30, str),
        # Todo: 13) Preencher com o conteúdo Do campo 15 da GNRE
    ]

class SintegraRegister74(SintegraRegister):
    sintegra_number = 74
    sintegra_fields = [
        ('inventory_date', 8, number),
        ('product_code', 14, str),
        ('product_quantity', 13, number),
        ('total_product_value', 13, number),
        # 'owner_code' accept the following values:
        # 1, Mercadorias de propriedade do Informante e em seu poder
        # 2, Mercadorias de propriedade do Informante em poder de terceiros
        # 3, Mercadorias de propriedade de terceiros em poder do Informante
        # according to the document linked in #3708
        ('owner_code', 1, number),
        ('owner_cnpj', 14, number),
        # The 'owner_state_registry' field must conform to the value in
        # 'owner_code' eg 'owner_state_registry' must be filled with the
        # company's data that really owns the product.
        ('owner_state_registry', 14, str),
        ('state', 2, str),
        ('blank', 45, str),
        ]
##
# 18. REGISTRO TIPO 70 - Nota Fiscal de Serviço de Transporte
#
class SintegraRegister70(SintegraRegister):
    sintegra_number = 70
    sintegra_fields = [
        ('cnpj', 14, number),
        # 2) CNPJ do emitente do documento, no caso de aquisição de serviço; CNPJ do tomador do serviço, no caso de
        #    emissão do documento
        ('estadual', 14, str),
        # 3) Inscrição Estadual do emitente do documento, no caso de aquisição de serviço; Inscrição Estadual do tomador
        #    do serviço, no caso de emissão do documento
        ('date', 8, number),
        # 4) Data de emissão para o prestador, ou data de utilização do serviço para o tomador
        ('state', 2, str),
        # 5) Sigla da unidade da Federação do emitente do documento, no caso de aquisição de serviço, ou do tomador do
        #    serviço, no caso de emissão do documento
        ('modelo', 2, number),
        # 6) Código do modelo da nota fiscal
        ('serie', 1, str),
        # 7) Série do documento
        ('subserie', 2, str),
        # 8) Subsérie do documento
        ('numero', 6, number),
        # 9) Número do documento
        ('cfop', 4, number),
        # 10) Código Fiscal de Operação e Prestação - Um registro para cada CFOP do documento fiscal
        ('valor', 13, number),
        # 11) Valor total do documento fiscal (com 2 decimais)
        ('icms_base', 14, number),
        # 12) Base de cálculo do ICMS (com duas decimais)
        ('icms', 14, number),
        # 13) Montante do imposto (com duas decimais)
        ('isento', 14, number),
        # 14) Valor amparado por isenção ou não incidência (com duas decimais)
        ('outras', 14, number),
        # 15) Valor que não confira débito ou crédito do ICMS (com duas decimais)
        ('modalidade', 1, number),
        # 16) Modalidade do frete:
        #     1-CIF;
        #     2-FOB;
        #     0-OUTR0S (a opção "0" - OUTROS nos casos em que não se aplica a informação de cláusula CIF ou FOB).
        ('situacao', 1, str),
        # 17) Situação do documento fiscal
    ]
##
# 19 - REGISTRO 71 (Informações da Carga Transportada)
#
class SintegraRegister71(SintegraRegister):
    sintegra_number = 71
    sintegra_fields = [
        ('cnpj', 14, number),
        # 2) CNPJ do tomador do serviço
        ('estadual', 14, str),
        # 3) Inscrição estadual do tomador do serviço
        ('date', 8, number),
        # 4) Data de emissão do conhecimento
        ('state', 2, str),
        # 5) Unidade da Federação do tomador do serviço
        ('modelo', 2, number),
        # 6) Modelo do conhecimento
        ('serie', 1, str),
        # 7) Série do conhecimento
        ('subserie', 2, str),
        # 8) Subsérie do conhecimento
        ('numero', 6, number),
        #  9) Número do documento
        ('state_remetente', 2, str),
        # 10) Unidade da Federação do remetente, se o destinatário for o tomador ou unidade da Federação do destinatário,
        #     se o remetente for o tomador
        ('cnpj_rem', 14, number),
        # 11) CNPJ do remetente, se o destinatário for o tomador ou CNPJ do destinatário, se o remetente for o tomador
        ('estadual_rem', 14, str),
        # 12) Inscrição Estadual do remetente, se o destinatário for o tomador ou Inscrição Estadual do destinatário,
        #     se o remetente for o tomador
        ('date_emit', 8, number),
        # 13) Data de emissão da nota fiscal que acoberta a carga transportada
        ('modelo_nf', 2, number),
        # 14) Modelo da nota fiscal que acoberta a carga transportada
        ('serie_nf', 3, str),
        # 15) Série da nota fiscal que acoberta a carga transportada
        ('numero_nf', 2, number),
        # 16) Número da nota fiscal que acoberta a carga transportada
        ('valor', 14, number),
        # 17) Valor total da nota fiscal que acoberta a carga transportada (com duas decimais)
        ('blank', 12, str)
    ]

class SintegraRegister75(SintegraRegister):
    sintegra_number = 75
    sintegra_fields = [
        ('start_date', 8, number),
        ('end_date', 8, number),
        ('product_code', 14, str),
        ('ncm', 8, str),
        ('descricao', 53, str),
        ('unit', 6, str),
        ('aliquota_ipi', 5, number),
        ('aliquota_icms', 4, number),
        ('reducao_icms', 5, number),
        ('base_icms', 13, number),
        ]


class SintegraRegister88EAN(SintegraRegister):
    sintegra_number = 88
    sintegra_fields = [
        ('subtipo', 3, str),
        ('versao_ean', 2, number),
        ('product_code', 14, str),
        ('descricao', 53, str),
        ('unit', 6, str),
        ('codigo_barra', 14, str),
        ('blank', 32, str)
        ]


class SintegraRegister90(SintegraRegister):
    sintegra_number = 90
    sintegra_fields = [
        ('cgc', 14, number),
        ('estadual', 14, str),
        ('type', 2, number),
        ('registers', 8, number),
        ('blank', 85, str),
        ('number', 1, number),
        ]
    sintegra_requires = 10, 11


def test():  # pragma nocover
    s = SintegraFile()
    cgc = int('03852995000107')
    estadual = '110042490114'

    s.add(SintegraRegister10(
        cgc, estadual, 'TESTES E TESTES LTDA',
        'CANDEIAS', 'SP', int('0710802316'),
        20070401, 20070430, '331'))

    s.add(SintegraRegister11(
        'RODOVIA BA 000 KM 00', 12, ',CX',
        'POSTAL 60', 43800000, 'EDSON / PEDRO',
        int('07100000000')))

    s.add_fiscal_coupon(datetime.date(2007, 4, 1),
                        '12345678901234567890', 1,
                        1, 10, 1, 1, 8,100)

    s.add_fiscal_coupon(datetime.date(2007, 4, 1),
        '12345678901234567890', 1,
        1, 10, 1, 1, 8,100)

    s.add(
        SintegraRegister60I('I', 20120102,'12365212541254563256','2D',456325,1,
            'COD01245',1,12356,125,'1800',245,' ' * 16,)
    )
    s.add(
        SintegraRegister51(12345678965458,'12456325632',20122563,'MG','12456321254562325621',
            1,1245,12563,125,0,0,'','N',))

    s.add(
        SintegraRegister53(12345678965458,'12456325632',20122563,'MG', 0o1,
            '123',1245,1363,'P',125,0,0,'N', 1, ''))
    s.add(
        SintegraRegister53(12345678965458,'12456325632',20122563,'MG', 0o1,
            '123',1245,1363,'P',125,0,0,'N', 1, ''))

    s.add(
        SintegraRegister55(12345678965458,'12456325632',20122563,'MG', 'SP', 123,
            3515,'AUT1234567896325412*',1363,20130102,1212,'conv'))
    s.add(
        SintegraRegister70(12345678965458,'12456325632', 20120203, 'MG', 0o1, 'U',
            'SS',123456, 1363, 1325, 1, 2, 3, 4, 1, 'N'))
    s.add(
        SintegraRegister71(12345678965458, '12456325632', 20120203, 'MG', 0o1, 'S', 'SB', 124523,
                           'SP', 12365254125489, '12354562', 20123501, 63, '325', 12341, 12, '')
    )


    s.add(
        SintegraRegister88('EAN',14,'20120203','PAO DE SAL','UN','012345678912345',32*' ')
    )


    s.close()
    s.write('/tmp/sintegra.txt')
    return s


if __name__ == '__main__':  # pragma nocover
    for s in test().get_registers():
        print('%r' % s.get_string())
